/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <cerrno>
#include <cmath>
#include <cassert>
#include <string>
#include <iostream>
#include <getopt.h>
#include "rumd/Sample.h"
#include "rumd/Timer.h"
#include "rumd/Device.h"
#include "rumd/IntegratorNVT.h"
#include "rumd/IntegratorIHS.h"
#include "rumd/RUMD_Error.h"
#include "rumd/PairPotential.h"

int main( int argc, char* argv[]) {
  try{
    int num_part = 1024;
    int parts_per_block = 16;
    int threads_per_part = 8;
    int NVT = 0;                         // 0=NVE 1=NVT (Nose-Hoover)
    int num_io_blocks = 1;
    int io_block_size = 1;
    int write_trajectory = 1;
    int dump_ihs = 1;                    // Quench every dump_ihs time steps.
    int momentum_reset_interval = 100;   // 0 to disable
    int ihs_iter = 1;                    // Number of iterations to do a quench.
    unsigned int quenchInterval = 50;
    //unsigned int lastQuench = 500;
    int mode = 1; // 0 = generate configuration, 1 = run trajectory and write final info; 2 = run trajectory and and compare final info to previous runs.
    
    float L=1.0;
    float Ti = 1.f;                      // Initial temperature
    float Tf = 1.f;                      // Final temperature
    float dt = 0.001f;
    float dt_ihs = 0.01f;
    float skin = 0.3f;
    float Rscal = 1.f;                   // Scaling of initial configuration
    float tau = 0.3f;
    //float4 sumU, sumK, sumW, sumS, sumC;
    //float stress[6];
    float totalDiffSq = 0.; // for mode 2
    
    bool Tf_assigned = false;
    
    while (1) {
      static struct option long_options[] = {
	{"n", 1, 0, 'n'},
	{"pb", 1, 0, 'p'},
	{"tp", 1, 0, 't'},
	{"thermostat", 1, 0, 'e'},
	{"numIOBlocks", 1, 0, 'o'},
	{"IOBlockSize", 1, 0, 'b'},
	{"writeTraj", 1, 0, 'w'},
	{"qi", 1, 0, 'q'},
	{"mode", 1, 0, 'm'},
	{"ihs_iter", 1, 0, 'c'},
	{"L", 1, 0, 'l'},
	{"Ti", 1, 0, 'i'},
	{"Tf", 1, 0, 'f'},
	{"dt", 1, 0, 'd'},
	{"tau", 1, 0, 'u'},
	{"skin", 1, 0, 's'},
	{"dump_ihs", 1, 0, 'a'},
	{"Rscal", 1, 0, 'r'},
	{0, 0, 0, 0}
      };
      int option_index = 0;
      int c = getopt_long(argc, argv, "",long_options, &option_index);
      if (c == -1)
	break;
      
      switch (c) {
      case 0:
	printf("option %s", long_options[option_index].name);
	if (optarg)
	  printf(" with arg %s", optarg);
        printf("\n");
	break;
      case 'n':
	num_part = atoi(optarg);
	break;
      case 'p':
	parts_per_block = atoi(optarg);
	break;
      case 't':
	threads_per_part = atoi(optarg);
	break;
      case 'e':
	NVT = atoi(optarg);
	break;
      case 'o':
	num_io_blocks = atoi(optarg);
	break;
      case 'a':
	dump_ihs = atoi(optarg);
	break;
      case 'b':
	io_block_size = atoi(optarg);
	break; 
      case 'c':
	ihs_iter = strtof(optarg, 0);
	break;
      case 'w':
	write_trajectory = atoi(optarg);
	break;
      case 'q':
	quenchInterval = (unsigned int)atoi(optarg);
	break;
      case 'm':
	mode = atoi(optarg);
      break;
      case 'l':
	L = strtof(optarg, 0);
	break;
      case 'i':
	Ti = strtof(optarg, 0);
	break;
      case 'f':
	Tf = strtof(optarg, 0);
	Tf_assigned = true;
	break;
      case 'd':
	dt = strtof(optarg, 0);
	break;
      case 'u':
	tau = strtof(optarg, 0);
	break;
      case 's':
	skin = strtof(optarg, 0);
	break;
      case 'r':
	Rscal = strtof(optarg, 0);
	break;
      case '?':
	exit(EXIT_FAILURE);
	break;
      default:
	printf("?? getopt returned character code 0%o ??\n", c);
      }
    }
    
    if (!Tf_assigned) Tf = Ti;
    
    Device::GetDevice().Init(argc, argv);
    
    // Create samples
    Sample* S = new Sample(parts_per_block, threads_per_part);
    Sample* SI = new Sample(parts_per_block, threads_per_part);
    
    // Create integrators.
    IntegratorNVT* itg =  new IntegratorNVT(dt, Ti);
    IntegratorIHS* itgI =  new IntegratorIHS(dt_ihs);
    
    // Setup the integrator parameters.
    itg->SetRelaxationTime(tau);
    itg->SetThermostatOn(NVT); 
    S->SetIntegrator(itg);
    SI->SetIntegrator(itgI);
    
    std::cout << "NVT pair potential" << std::endl;
    Pot_LJ_12_6_Gromacs potential;   
    potential.SetParams(0, 0, 1.00, 1.00, 2.5);
    potential.SetParams(0, 1, 0.80, 1.50, 2.5);
    potential.SetParams(1, 0, 0.80, 1.50, 2.5);
    potential.SetParams(1, 1, 0.88, 0.50, 2.5);
    potential.SetNbListSkin(skin);
    
    std::cout << "IHS pair potential" << std::endl;
    Pot_LJ_12_6_Gromacs potentialI;   
    potentialI.SetParams(0, 0, 1.00, 1.00, 2.5);
    potentialI.SetParams(0, 1, 0.80, 1.50, 2.5);
    potentialI.SetParams(1, 0, 0.80, 1.50, 2.5);
    potentialI.SetParams(1, 1, 0.88, 0.50, 2.5);
    potentialI.SetNbListSkin(skin);
    S->SetPotential(&potential);
    SI->SetPotential(&potentialI);
    
    potential.WritePotentials(S->GetSimulationBox());
    //S->DumpPotentials();
    
    if(mode == 0){
      throw RUMD_Error("ConsistencyTest","main", "mode = 0 has been disabled");
      printf("Creating random conf\n");
      //S->CreateRandomConf(num_part, L, 0.8, 20*Ti);
      S->WriteConf("new_conf.xyz"); // different name from start, must be renamed (to avoid accidental overwriting)
    }
    else{
      S->ReadConf("start.xyz.gz");
      SI->ReadConf("start.xyz.gz");     
      itg->SetMomentumToZero();
    }
    
    // Calculate forces
    
    /*
    // Perform a single quench.
    *SI = *S;
    for( int j=0; j < ihs_iter; j++ ){
    SI->CalcF(); 
    itgI->Integrate();
    }
    
    // Set the initial config for IHS comparison.
    //itgI->SetPreviousInherentStateConfiguration();
    */
    
    S->CalcF(); 
    //S->CalcEnergies(&sumU, &sumK, &sumW, &sumS, &sumC, stress);
    //float Tt = sumK.y;
    unsigned numParticles = S->GetNumberOfParticles();
    double pe_init = S->GetPotentialEnergy();
    double ke_init = S->GetIntegrator()->GetKineticEnergy();
    double vir_init = S->GetVirial();
    printf("Have initial conf; initial potential energy is %f\n",pe_init/numParticles);
    double startU_diff = -1.0, startK_diff = -1.0, startW_diff = -1.0, totalE=-1.0;
    FILE *start_energy_f;
    if(mode == 1) {
      start_energy_f = fopen("start_energies.dat","w");
      if(!start_energy_f){
	fprintf (stderr,"Error opening file start_energies.dat: %s\n", strerror(errno));
	exit(1);
      }
      fprintf(start_energy_f, "%.8f %.8f %.8f %.8f\n",  pe_init, ke_init, 
	    pe_init + ke_init, vir_init);
      fclose(start_energy_f);
    }
    else if (mode == 2){
      start_energy_f = fopen("start_energies.dat","r");
      if(!start_energy_f){
	fprintf (stderr,"Error opening file start_energies.dat: %s\n", strerror(errno));
	exit(1);
      }
      
      assert(start_energy_f);
      int f_stat = fscanf(start_energy_f, "%lf %lf %lf %lf", &startU_diff, &startK_diff, &totalE, &startW_diff);
      assert(f_stat == 4);
      startU_diff -= pe_init;
      startK_diff -= ke_init;
      startW_diff -= vir_init;
      startU_diff /= numParticles;
      startK_diff /= numParticles;
      startW_diff /= numParticles;
      totalDiffSq += startU_diff*startU_diff;
      totalDiffSq += startK_diff*startK_diff;
      totalDiffSq += startW_diff*startW_diff;
      fclose(start_energy_f);  
    }
        
    // Setup of various stuff
    S->SetOutputBlockSize(io_block_size);
    S->SetWriteTrajectory(true);
    S->GetTrajectoryConfMetaData().has_pot_energies = true;
    S->GetTrajectoryConfMetaData().has_virials = true;
    S->GetTrajectoryConfMetaData().precision = 9;
    S->SetEnergyLinearWriteInterval(1);
    S->GetEnergiesMetaData().precision = 9;
    S->GetEnergiesMetaData().Set("thermostat_Ps", true);

    Timer timer;
    // double eps = 1.e-10;
    
    unsigned long int num_iter = (unsigned long int)num_io_blocks * io_block_size;
    
    printf("Doing %lu timesteps:\n", num_iter);

    FILE* IHSData; IHSData = fopen("ihs_sampled.dat", "w");
    
    for( unsigned long int i = 0; i < num_iter; i++) {
      
      //if(mode == 0 && !NVT && i%quenchInterval == 0 && i < lastQuench)
      //  S->ZeroAllVelocities();  
      
      // Calculate forces
      S->CalcF();
      S->NotifyOutputManagers(i);
      
      // Copies the minimum GPU sample for integration to continue.
      if( dump_ihs > 0 && i % dump_ihs == 0 ){
	*SI = *S; 
	
	for( int j=0; j < ihs_iter; j++ ){
	  SI->CalcF(); 
	  itgI->Integrate();
	}
	
	// Detect a transition between inherent states.
	//float length = itgI->GetInherentStateTransitionLength();
	//if( length > 0.1f ){
	//itgI->DumpInherentStateTransitionConfigurations();      
	//itgI->SetPreviousInherentStateConfiguration();
	//}
	fprintf(IHSData, "%e %e \n", itgI->GetInherentStatePotentialEnergy(), itgI->GetInherentStateForceSquared());
      } 
      
      // Integrate velocities and positions
      itg->Integrate();

      if ( momentum_reset_interval > 0 && i % momentum_reset_interval==0 ) 
	itg->SetMomentumToZero();
    }	

    fclose(IHSData);
    system("mv ihs_sampled.dat TrajectoryFiles/");

    double gpu_time = timer.elapsed();
    printf( "\nProcessing time GPU (s): %f\n", gpu_time);
    printf( " per timestep  GPU (ms): %f\n", 1000.0*gpu_time/num_iter);
    
    // Update forces and velocities
    S->CalcF();
    //S->CalcEnergies(&sumU, &sumK, &sumW, &sumS, &sumC, stress);
    double pe_final = S->GetPotentialEnergy();
    double ke_final = S->GetIntegrator()->GetKineticEnergy();
    double vir_final = S->GetVirial();
    //Tt = sumK.y;
    double endU_diff = 0., endK_diff = 0., endW_diff = 0.;
    if (mode == 0) {
      S->WriteConf("start.xyz");
    }
    if(mode == 1) {
      FILE *end_energy_f = fopen("end_energies.dat","w");
      if(!end_energy_f){
	fprintf (stderr,"Error opening file end_energies.dat: %s\n", strerror(errno));
	exit(1);
      }
      
      fprintf(end_energy_f, "%.8f %.8f %.8f %.8f %.8f\n", pe_final, ke_final, pe_final+ke_final, vir_final, gpu_time);
      fprintf(end_energy_f, "nSteps= %lu; dt=%f\n",num_iter, itg->GetTimeStep());
      fclose(end_energy_f);
      S->WriteConf("end.xyz");
    }
    else if(mode == 2) {
      printf("mode 2\n");
      FILE *end_energy_f = fopen("end_energies.dat","r");
      if(!end_energy_f){
	fprintf (stderr,"Error opening file end_energies.dat: %s\n", strerror(errno));
	exit(1);
      }
      
      double totalE, prevGpuTime;
      fscanf(end_energy_f, "%lf %lf %lf %lf %lf", &endU_diff, &endK_diff, &totalE, &endW_diff, &prevGpuTime);
      endU_diff -= pe_final;
      endK_diff -= ke_final;
      endW_diff -= vir_final;
      endU_diff /= numParticles;
      endK_diff /= numParticles;
      endW_diff /= numParticles;
      
      totalDiffSq += endU_diff*endU_diff;
      totalDiffSq += endK_diff*endK_diff;
      totalDiffSq += endW_diff*endW_diff;
      FILE *statusFile = fopen("status.txt","w");
      if(!statusFile){
	fprintf (stderr,"Error opening file status.txt: %s\n", strerror(errno));
	exit(1);
      }
	  
      fprintf(statusFile, "Status for N=%d, L=%.2f, pb=%d, tp=%d, ",numParticles,L,parts_per_block,threads_per_part);
      if(NVT)
	fprintf(statusFile, "NVT : ");
      else
	fprintf(statusFile, "NVE : ");
      if(totalDiffSq < 1.e-6)
	// need to allow for changes in round-off error when different
	// (pb, tp) combinations are used so tolerance should not be too small
	fprintf(statusFile, "OK\n");
      else
	fprintf(statusFile, "FAIL\n");
      
      fprintf(statusFile,"Total difference squared: %g\n",totalDiffSq);
      fprintf(statusFile,"startU_diff: %.8f\n",startU_diff);
      fprintf(statusFile,"startK_diff: %.8f\n",startK_diff);
      fprintf(statusFile,"startW_diff: %.8f\n",startW_diff);
      fprintf(statusFile,"endU_diff: %.8f\n",endU_diff);
      fprintf(statusFile,"endK_diff: %.8f\n",endK_diff);
      fprintf(statusFile,"endW_diff: %.8f\n",endW_diff);
      fprintf(statusFile,"time taken, before/now (s): %f/%f\n",prevGpuTime, gpu_time);
    }
    delete S;
    delete SI;
  } 
  catch(RUMD_Error e){
    std::cerr << "RUMD_Error thrown from function " << e.className << "::" << e.methodName << ": " << std::endl << e.errorStr;
    return 1;
  }
  return 0;
}
