/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <string>
#include <getopt.h>
#include <iostream>

#include "rumd/Device.h"
#include "rumd/Timer.h"
#include "rumd/Sample.h"
#include "rumd/RUMD_Error.h"
#include "rumd/IntegratorMMC.h"
#include "rumd/PairPotential.h"

int main( int argc, char* argv[]) {
  try{
    int parts_per_block = 16;
    int threads_per_part = 8;
    int num_io_blocks = 1;               // Number of (logarthmic) configuration blocks
    int io_block_size = 1;
    int restartBlock = -1;               // which block to restart from (-1 means no restart)
    int energyWriteInterval = 200;       // Interval (timesteps) between dumping of energies
    
    float Ti = 1.f;                      // Initial temperature
    float Tf = 1.f;                      // Final temperature
    float dt = 0.0025f;
    float skin = 0.3f;
    float Rscal = 1.f;                   // Scaling of initial configuration
    bool writeTrajectory = true;
    //bool writeEnergies = true;
    bool Tf_assigned = false;
    
    while (1) {
      static struct option long_options[] = {
	{"pb", 1, 0, 'p'},
	{"tp", 1, 0, 't'},
	{"numIOBlocks", 1, 0, 'o'},
	{"IOBlockSize", 1, 0, 'b'},
	{"dump_nrg", 1, 0, 'w'},
	{"restartBlock", 1, 0, 'x'},
	{"Ti", 1, 0, 'i'},
	{"Tf", 1, 0, 'f'},
	{"dt", 1, 0, 'd'},
	{"skin", 1, 0, 's'},
	{"Rscal", 1, 0, 'r'},
	{"writeTrajectory", 1, 0, 'y'},
	//{"writeEnergies",1,0,'g'},
	{0, 0, 0, 0}
      };
      int option_index = 0;
      int c = getopt_long(argc, argv, "",long_options, &option_index);
      if (c == -1)
	break;
      
      switch (c) {
      case 0:
	printf("option %s", long_options[option_index].name);
	if (optarg)
	  printf(" with arg %s", optarg);
        printf("\n");
	break;
      case 'p':
	parts_per_block = atoi(optarg);
	break;
      case 't':
	threads_per_part = atoi(optarg);
	break;
      case 'o':
	num_io_blocks = atoi(optarg);
	break;
      case 'b':
	io_block_size = atoi(optarg);
	break; 
      case 'w':
	energyWriteInterval = atoi(optarg);
	break;
      case 'x':
	restartBlock = atoi(optarg);
	break;
      case 'i':
	Ti = strtof(optarg, 0);
	break;
      case 'f':
	Tf = strtof(optarg, 0);
	Tf_assigned = true;
	break;
      case 'd':
	dt = strtof(optarg, 0);
	break;
      case 's':
	skin = strtof(optarg, 0);
	break;
      case 'r':
	Rscal = strtof(optarg, 0);
	break;
      case 'y':
	writeTrajectory = (bool) atoi(optarg);
	break;
	//case 'g':
	//writeEnergies = (bool) atoi(optarg);
	//break;
      case '?':
	exit(EXIT_FAILURE);
	break;
      default:
	printf("?? getopt returned character code 0%o ??\n", c);
      }
    }
    
    if (!Tf_assigned) Tf = Ti;
    
    Device::GetDevice().Init(argc, argv);    

    // Create a new sample
    Sample* S = new Sample(parts_per_block, threads_per_part);

    // Create an integrator.
    IntegratorMMC* itg =  new IntegratorMMC();
    
    // Setup the integrator parameters.
    itg->SetTimeStep(dt);
    itg->SetTargetTemperature(Ti);
    S->SetIntegrator(itg);
        
    // Create LJ pair-potential
    Pot_LJ_12_6_Gromacs potential;   
    
    // Setup parameters for Lennard-Jones interactions ( Kob & Andersen )
    // i  j   Sig   Eps  Rcut (in units of Sigma_ij)
    potential.SetParams(0, 0, 1.00, 1.00, 2.5);
    potential.SetParams(0, 1, 0.80, 1.50, 2.5);
    potential.SetParams(1, 0, 0.80, 1.50, 2.5);
    potential.SetParams(1, 1, 0.88, 0.50, 2.5);
    potential.SetNbListSkin(skin);
    S->SetPotential(&potential);
    
    // Read configuration & Copy it to device
    unsigned long int initialTimeStep = 0;
    
    if(restartBlock < 0) {
      S->ReadConf("start.xyz.gz");
      S->IsotropicScaleSystem(Rscal);
    }  
    else {
      S->ReadRestartConf(restartBlock);
      initialTimeStep = (unsigned long int) (restartBlock)*io_block_size;
    }
     
    potential.WritePotentials(S->GetSimulationBox());
    
    // Setup of various stuff
    S->SetOutputBlockSize(io_block_size);
    S->SetWriteTrajectory(writeTrajectory);
    if(energyWriteInterval) {
      S->SetWriteEnergies(true);
      S->SetEnergyLinearWriteInterval(energyWriteInterval);
    }
    S->GetTrajectoryConfMetaData().precision = 9;
    S->GetTrajectoryConfMetaData().has_pot_energies = true;
    S->GetTrajectoryConfMetaData().has_virials = true;
        
    S->GetEnergiesMetaData().precision = 9;
    S->GetEnergiesMetaData().Set("configTemperature", true);
    S->GetEnergiesMetaData().Set("laplacian", true); 
    
    if(num_io_blocks > 9999){
      fprintf(stderr, "Error: Number of IO blocks cannot be more than 9999.\n");
      exit(1);
    }
    
    unsigned long int num_iter = (unsigned long int) num_io_blocks*io_block_size + 1;
            
    printf("Total number of timestep(s) is %lu, ", num_iter);
    if(restartBlock >= 0)
      printf("restart: starting from iteration %lu\n",initialTimeStep);
    printf("dumping energies every %d timestep (every %f time-unit).\n", energyWriteInterval, energyWriteInterval*itg->GetTimeStep());
    
    // Execute the kernel(s)
    Timer timer;
    
    for( unsigned long int i = initialTimeStep; i < num_iter; i++ ){
      itg->Integrate(); 
      S->CalcF();
      S->NotifyOutputManagers(i); 
    }
    
    double gpu_time = timer.elapsed();
    printf( "\nProcessing time GPU (s): %f\n", gpu_time);
    printf( " N | thermostat | pb | tp | ms/timestep | accept/timestep \n");
    float t_per_step = 0.;
    if (num_iter > 0) t_per_step = 1000.0*gpu_time/num_iter;
    printf( "%d | %d | %d | %f | %f\n", S->GetNumberOfParticles(), parts_per_block, threads_per_part, t_per_step, itg->GetAcceptanceRate());
    
    // Update forces and velocities and copy everything back from device 
    S->CalcF();  
    S->WriteConf("confout.xyz");
    
    // Free objects.
    delete S; 
    delete itg;
  }
  catch(RUMD_Error e){
    std::cerr << "RUMD_Error thrown from function " << e.className << "::" << e.methodName << ": " << std::endl << e.errorStr << std::endl;
    return 1;
  }
  return 0;
}
