#ifndef SAMPLE_H
#define SAMPLE_H

/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <vector>

#include "rumd/rumd_base.h"
#include "rumd/rumd_technical.h"
#include "rumd/ConfigurationWriterReader.h"
#include "rumd/LogLinOutputManager.h"
#include "rumd/EnergiesOutputManager.h"
#include "rumd/EnergiesMetaData.h"

#include "rumd/SimulationBox.h"
#include "rumd/ParticleData.h"
#include "rumd/NeighborList.h"
#include "rumd/KernelPlan.h"

class Integrator; class Potential; 

class Sample{
  friend class ConfigurationWriterReader;
  friend class MoleculeData;
  friend class BondHarmonic;
  friend class BondFENE;
  friend class ConstraintPotential;

 private:
  unsigned int particles_per_block;
  unsigned int threads_per_particle;
  unsigned int degreesOfFreedom;
  unsigned long int blockSize;

  std::string trajectoryDir;
  bool writers_initialized;
  std::vector<LogLinOutputManager*> outputManagers;

  LogLinOutputManager* restartLogLin;
  LogLinOutputManager* trajectoryLogLin;
  EnergiesOutputManager* energiesLogLin; 
  std::vector<Potential*> potentialList;

  // Input/Output
  bool verbose;
  float4 CalculateCenterOfMass( float4* position, float4* velocity, int moleculeIndex );

  ConfigurationWriterReader* restartWriter;
  ConfigurationWriterReader* trajectoryWriter;
  void CreateDefaultOutputManagers();

 protected:
  SimulationBox* simulationBox;
  Integrator* itg; 
  ParticleData particleData;
  KernelPlan kPlan;  
  
  unsigned int numTypes;
  float massOfType[MaxNumTypes];
  
 public:
  Sample( unsigned int pb, unsigned int tp );
  Sample(const Sample& S);
  ~Sample();
  Sample& operator=(const Sample& S); 
      
  void InitializeOutputManagers(unsigned long int timeStepIndex);
  void NotifyOutputManagers(unsigned long int timeStepIndex);
  void TerminateOutputManagers(); // close open files etc
  void ReadConf( const std::string& filename );
  void ReadRestartConf(int restartBlock, unsigned int numDigitsBlkIdx=4 );
  void WriteConf(const std::string& filename, const std::string& mode="w"){ restartWriter->Write(this, filename, mode); }
  void AddOutputManager(LogLinOutputManager* om);
 
  ConfigurationMetaData& GetTrajectoryConfMetaData() {return trajectoryWriter->metaData; }
  EnergiesMetaData& GetEnergiesMetaData() {return energiesLogLin->metaData;}
  EnergiesOutputManager& GetEnergiesOutputManager() {return (*energiesLogLin);}

  void CalcF(bool calc_stresses = false, bool calc_temp = false, bool calc_bs = false);
  void IsotropicScaleSystem( float Rscal );
  void CopySimulationDataToDevice();
  void CopySimulationDataFromDevice();
  void AddPotential(Potential* potential);
  void SubtractDOFs( unsigned i ){ degreesOfFreedom -= i; };

  // Set methods
  void SetOutputBlockSize(unsigned long int blockSize);
  void SetWriteEnergies(bool writeEnergies);
  void SetWriteTrajectory(bool writeTrajectory);
  void SetWriteRestart(bool writeRestart);
  void SetTrajectoryLinearWriteInterval(unsigned long int writeInterval);
  void SetTrajectoryLogLinParams(unsigned int base, unsigned int maxInterval);
  void SetEnergyLinearWriteInterval(unsigned int writeInterval); 
  void SetEnergyLogLinParams(unsigned int base, unsigned int maxInterval); 
  void SetNumberOfParticles(unsigned int num_part);
  void SetSimulationBox(SimulationBox* simBox){ delete simulationBox; simulationBox = simBox; };
  void SetIntegrator(Integrator* newItg);
  void SetPotential(Potential* potential);
  void SetVerbose(bool vb);
  void SetOutputDirectory(const std::string& trajectoryDir);

  // Get methods.
  const ParticleData& GetParticleData() const { return particleData; }
  const KernelPlan& GetKernelPlan() const { return kPlan; }
  const Integrator* GetIntegrator() const { return itg; }
  SimulationBox* GetSimulationBox() const { return simulationBox; }  
  std::vector<Potential*>* GetPotentials() { return &potentialList; }  

  unsigned int GetParticlesPerBlock() const { return particles_per_block; }
  unsigned int GetThreadsPerParticle() const { return threads_per_particle; }
  unsigned int GetNumberOfDOFs() const { return degreesOfFreedom; }
  unsigned int GetNumberOfParticles() const { return particleData.GetNumberOfParticles(); }
  unsigned int GetNumberOfVirtualParticles() const { return particleData.GetNumberOfVirtualParticles(); }
  unsigned int GetNumberOfTypes() const { return numTypes; }
  float GetMass(unsigned int type) const {return massOfType[type];}

  double GetPotentialEnergy() const;  
  double GetVirial() const;
  const std::string& GetOutputDirectory() const { return trajectoryDir; }
  unsigned long int GetOutputBlockSize() const { return blockSize; }
};

#endif // SAMPLE_H
