
/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include <fstream>
#include <typeinfo>

#include "rumd/rumd_base.h"
#include "rumd/rumd_utils.h"
#include "rumd/PairPotential.h"
#include "rumd/NeighborList.h"

#include <sstream>
#include <iostream>
__device__ __constant__ float d_params[MaxNumPP_Objects][MaxNumTypes][MaxNumTypes][NumParam];

template<int STR, int CONFT, class P, class S> 
__device__ __host__ float fij( P* Pot, float4 my_r, float4 rj, float4* my_f, float4* my_w, 
			       float4* my_sts, float4* my_misc, float* param, 
			       S* simBox, float* simBoxPointer );


template<int STR, int CONFT, bool initialize, class P, class S>
__device__ void calc_f( P *Pot, float4 my_r, float4 *r, float4 *f, float4 *w, 
			float4 *sts, float4* misc, unsigned int my_type, S* simBox, 
			float* simBoxPointer, NBTYPEv *NbList, float pot_factor_over2, unsigned int objectIdx );

template<int STR, int CONFT, bool initialize, class P, class S> 
__global__ void md_calc_f( P *Pot, float4* r, float4* f, float4* w, float4* sts, 
			   float4* misc, S* simBox, float* simBoxPointer, bool isCubic, 
			   NBTYPEv* NbList, float pot_factor_over2, unsigned int objectIdx );

__device__ float4 Plus_Float4( const float4 &a, const float4 &b){ 
  float4 c = {a.x + b.x, a.y + b.y, a.z + b.z, a.w + b.w}; return c;
}


using namespace std;

//unsigned int PairPotential::pp_objectCounter = 0;
std::vector<unsigned int> PairPotential::objectIndexList;

//////////////////////////////////////////////////
// Abstract PairPotential implementations.
//////////////////////////////////////////////////

void PairPotential::CopyParamsToGPU() {
  // Copy pair-interaction parameters to Device
  if(verbose)
    cout << "Copying pair-interaction parameters to GPU (index= " << pp_objectIndex  << ", label=" << GetID_String() << ")" << endl;
  unsigned int size = sizeof(float) * MaxNumTypes * MaxNumTypes * NumParam;
  // need to use the "offset" argument to cudaMemcpyToSymbol ... doing
  // pointer arithmetic on the first argument doesn't work
  cudaMemcpyToSymbol(d_params, h_params, size, pp_objectIndex * size);
} 

void PairPotential::WritePotentials(SimulationBox* simBox){
  RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(simBox);
  
  float4 my_r = {0.0f, 0.0f, 0.0f, 0.0f};
  float4    r = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_f = {0.0f, 0.0f, 0.0f, 0.0f}; 
  float4 my_w = {0.0f, 0.0f, 0.0f, 0.0f}; 

  ofstream pFile( (string("potentials_") + GetID_String() + ".dat").c_str() );
  
  for (unsigned int i=1; i < 999; i++){
    // Update only the x component.
    my_r.x = ( i * powf(simBox->GetVolume(), 1.f/3.f) ) / float(1000);
    pFile << my_r.x << " ";
    
    for(unsigned int MyType=0; MyType < MaxNumTypes; MyType++){
      for(unsigned int Type=0; Type < MaxNumTypes; Type++){
	my_f.w = 0.0f; 
	
	float* param = h_params[MyType][Type];
	float4 distance = testRSB->calculateDistance( my_r, r, testRSB->GetHostPointer() );
				   						     
	if( distance.w <= param[0] && distance.w >= 0.000001f ){
	  float s = ComputeInteraction_cpu(distance.w, param, &my_f, &my_w);
      	  my_f.x += distance.x * s;
	  my_f.y += distance.y * s;
	  my_f.z += distance.z * s;
      	}
	pFile << my_f.w * potential_factor << " ";
      }
    }
    pFile << endl;
  } 
  pFile.close();
}

void PairPotential::AllocatePE_Array(unsigned int nvp) {
  if(allocatedSize != 0) {
    cudaFreeHost(h_f_pe);
    cudaFree(d_f_pe);
  }

   if( cudaMallocHost( (void**) &h_f_pe, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation ) 
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on h_f") );
    if( cudaMalloc( (void**) &d_f_pe, nvp * sizeof(float4) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("ParticleData","AllocateParticles","Malloc failed on d_f") );
    memset( h_f_pe, 0, nvp * sizeof(float4) );
    cudaMemset( d_f_pe, 0, nvp * sizeof(float4) );

    allocatedSize = nvp;
}

double PairPotential::GetPotentialEnergy(){
  neighborList.UpdateNBlist(simBox);

  if(allocatedSize != particleData->GetNumberOfVirtualParticles())
    AllocatePE_Array(particleData->GetNumberOfVirtualParticles());

  // launch kernel with d_f_pe instead of df as the force array
  LaunchPE_Kernel();

  double pe_thisPotential = 0.;
  // copy to host
  cudaMemcpy( h_f_pe, d_f_pe, allocatedSize * sizeof(float4), cudaMemcpyDeviceToHost );
  // sum over particles

  for(unsigned int i=0; i < nParticles; i++)
    pe_thisPotential += h_f_pe[i].w;

  return pe_thisPotential;
}

void PairPotential::SetSampleParameters(KernelPlan& kPlan,  ParticleData* pd, SimulationBox* simBox, float4* d_misc){
  Potential::SetSampleParameters(kPlan, pd, simBox, d_misc);
  neighborList.Initialize(pd, kPlan);
  neighborList.SetCutoffs(this);

  testRSB = dynamic_cast<RectangularSimulationBox*>(simBox);

  CopyParamsToGPU();
}

void PairPotential::SetExclusionBond(int1 *h_btlist, uint1* h_blist, unsigned num_part, 
				     unsigned num_virt_part, unsigned max_num_bonds, 
				     unsigned etype ){
  neighborList.SetExclusionBond(h_btlist, h_blist, num_part, num_virt_part, max_num_bonds,  etype);
}

void PairPotential::SetExclusionMolecule(int1 *h_mlist, unsigned molindex, unsigned max_num_uau){
  neighborList.SetExclusionMolecule(h_mlist, molindex, max_num_uau);
}

/////////////////////////////////////////////////////////
// Specialized PairPotentials implementations.
// The specialization is needed due to CUDA limitations.
/////////////////////////////////////////////////////////

void Pot_LJ_12_6::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs) {
  neighborList.UpdateNBlist(simBox);

  if(initialize){
    if(calc_stresses)
      md_calc_f<1,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex ); 
    else
      md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
}

void Pot_LJ_12_6::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}



void Pot_gLJ_m_n::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs) {
  neighborList.UpdateNBlist(simBox);

  if(initialize){
    if(calc_stresses)
      md_calc_f<1,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
}

void Pot_gLJ_m_n::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}

void Pot_LJ_12_6_NEW::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs) {
  neighborList.UpdateNBlist(simBox);
  
  if(initialize){
    if(calc_stresses)
      md_calc_f<1,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
} 

void Pot_LJ_12_6_NEW::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}



void Pot_LJ_12_6_Gauss::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs)
{
  neighborList.UpdateNBlist(simBox);
  
  if(initialize){
    if(calc_stresses)
      md_calc_f<1,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
} 

void Pot_LJ_12_6_Gauss::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}


void  Pot_Gauss::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs) {
  neighborList.UpdateNBlist(simBox);
 
  if(initialize){
    if(calc_stresses)
      md_calc_f<1,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
}


void Pot_Gauss::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}


void Pot_Buckingham::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs) {
  neighborList.UpdateNBlist(simBox);

  if(initialize){
    if(calc_stresses)
      md_calc_f<1,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
}  

void Pot_Buckingham::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}


void Pot_IPL_12::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs) {
  neighborList.UpdateNBlist(simBox);

  if(initialize){
    if(calc_stresses)
      md_calc_f<1,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
} 


void Pot_IPL_12::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}

void Pot_IPL_18::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs) {
  neighborList.UpdateNBlist(simBox);

  if(initialize){
    if(calc_stresses)
      md_calc_f<1,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
} 

void Pot_IPL_18::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}

void Pot_IPL_n::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs) {
  neighborList.UpdateNBlist(simBox);

  if(initialize){
    if(calc_stresses)
      md_calc_f<1,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,0,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
}  

void Pot_IPL_n::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}

void Pot_LJ_12_6_Gromacs::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs) { // Config temp is currently always on.
  neighborList.UpdateNBlist(simBox);
  
  if(initialize){
    if(calc_stresses)
      md_calc_f<1,1,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,1,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,1,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,1,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
}

void Pot_LJ_12_6_Gromacs::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}


void Pot_Dzugutov::CalcF(bool initialize, bool calc_stresses, bool calc_temp, bool calc_bs) { // Config temp is currently always on.
  neighborList.UpdateNBlist(simBox);
  
  if(initialize){
    if(calc_stresses)
      md_calc_f<1,1,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,1,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
  else{
    if(calc_stresses)
      md_calc_f<1,1,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
    else
      md_calc_f<0,1,0><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), 
								   testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
  }
}

void Pot_Dzugutov::LaunchPE_Kernel() {
  md_calc_f<0,0,1><<< kp.grid, kp.threads, kp.shared_size >>>( this, d_r, d_f_pe, d_w, d_sts, d_misc, testRSB, testRSB->GetDevicePointer(), testRSB->GetIsCubicShape(), neighborList.NbList, potential_factor*0.5, pp_objectIndex );
}



/////////////////////////////////////////////////////////
// Kernel implementation of pair force evaluation.
/////////////////////////////////////////////////////////

template<int STR, int CONFT, class P, class S> 
__device__ __host__ float fij( P* Pot, float4 my_r, float4 rj, float4* my_f, float4* my_w, float4* my_sts, float4* my_misc, float* param, S* simBox, float* simBoxPointer ){

  float4 dist = simBox->calculateDistance(my_r, rj, simBoxPointer);

  // Inside cut-off for interaction? (param[0]=Rcut^2)
  if ( dist.w <= param[0] && dist.w >= 0.000001f ){
    float s;
    
    if(CONFT){
      s = Pot->ComputeInteraction(dist.w, param, my_f, my_w, my_misc);
    }
    else{
      s = Pot->ComputeInteraction(dist.w, param, my_f, my_w);
    }
    
    (*my_f).x += dist.x * s;
    (*my_f).y += dist.y * s;
    (*my_f).z += dist.z * s;
    
    if(STR){
      // stress - diagonal components
      (*my_sts).x -= dist.x * dist.x * s;  // xx
      (*my_sts).y -= dist.y * dist.y * s;  // yy
      (*my_sts).z -= dist.z * dist.z * s;  // zz
      // stress - off-diagonal components
      (*my_sts).w -=  dist.y * dist.z * s; // yz
      (*my_w).y -=  dist.x * dist.z * s;   // xz
      (*my_w).z -=  dist.x * dist.y * s;   // xy      
    }
  } // END  dist.w <= param[0]...
  
  return dist.w;                                                 
}

template<int STR, int CONFT, bool initialize, class P, class S> 
__device__ void calc_f( P* Pot, float4 my_r, float4* r, float4* f, float4* w, 
			float4* sts, float4* misc, unsigned int my_type, S* simBox, 
			float* simBoxPointer, NBTYPEv* NbList, float pot_factor_over2, unsigned int objectIdx ){
  extern __shared__ float4 s_r[];

  float4 total_f = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 total_w = {0.0f, 0.0f, 0.0f, 0.0f};  
  float4 total_sts = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 total_misc = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_f = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_w = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_sts = {0.0f, 0.0f, 0.0f, 0.0f};
  float4 my_misc = {0.0f, 0.0f, 0.0f, 0.0f};

  if(!initialize){ // both sts and misc shoud follow, but currently not being intialized elsewhere?
    total_w = w[MyGP];
    total_f = f[MyGP]; 

    if(STR)
      total_sts = sts[MyGP];
    
    if(CONFT)
      total_misc = misc[MyGP];
  }

  const unsigned int offset = MyT*PPerBlock;           
  const unsigned int CellOffset = MyB*NumBlocks*PPerBlock; 
  unsigned int Pread, j, cum_j;
  NBTYPE NbListCell;
  
  // Loop over all cells
  for( unsigned int cell0 = 0; cell0 < NumBlocks; cell0+=TPerPart ){
    unsigned int cell = cell0 + MyT;
    
    Pread = __umul24(cell,PPerBlock) + MyP;
    
    // Read coordinates from global to shared mem. NOTE: must be same method as in nb-list!!!
    if ( cell < NumBlocks ){
      s_r[MyP+offset] = r[Pread]; 
    }
    
    __syncthreads();  
    
    if (cell<NumBlocks){
      // Read the NbList (coded as a binary mask) from global memory to register 
      NbListCell = NbList[CellOffset + Pread].x;
      
      // Run through NbList (i.e. the '1'-bits in NbListCell)   
      cum_j = 0; 
      
      while (NbListCell>0) {
	cum_j += (j = __ffs(NbListCell));         // find next '1'-bit (from the right, labeling bit 0 as 1) 
	NbListCell = (NbListCell >> 1) >> j-1;    // remove the '1'-bit we just found
	
	int index = offset + cum_j - 1; 
	
	fij<STR, CONFT>( Pot, my_r, s_r[index], &my_f, &my_w, &my_sts, &my_misc, d_params[objectIdx][my_type][__float_as_int(s_r[index].w)], simBox, simBoxPointer );
      }  
    } // END if (cell<NumBlocks)
    __syncthreads();  
    
  } // END for( unsigned int cell0 = 0; ... 
  
  // Sum the force for all local threads. 'Borrowing s_r'
  s_r[MyP+offset] = my_f;
  
  __syncthreads();  
  
  if( MyT == 0 ){
    for( int i=1; i < TPerPart; i++ ){
      int j = MyP + __umul24(i,PPerBlock); 
      my_f.x += s_r[j].x;
      my_f.y += s_r[j].y;
      my_f.z += s_r[j].z;
      my_f.w += s_r[j].w;
    }

    my_f.w *= pot_factor_over2;

    if(initialize)
      f[MyGP] = my_f;
    else
      f[MyGP] = Plus_Float4(my_f, total_f); // do the addition locally before writing
  }
  
  __syncthreads();  
  // Sum the atomic/molecular virial for all local threads.
  s_r[MyP+offset] = my_w;
  __syncthreads();  
  
  if( MyT == 0 ){
    for ( int i=1; i < TPerPart; i++ ){
      int j = MyP + __umul24(i,PPerBlock); 
      my_w.w += s_r[j].w;

      if(STR){
	my_w.y += s_r[j].y;
	my_w.z += s_r[j].z;
      }
    }

    if(initialize)
      w[MyGP] = my_w;
    else
      w[MyGP] = Plus_Float4(my_w, total_w);
  }
  
  if(STR){
    __syncthreads();  
    // Sum the shear stress for all local threads.
    s_r[MyP+offset] = my_sts;
    __syncthreads();  
    
    if( MyT == 0 ){
      for ( int i=1; i < TPerPart; i++ ){
	int j = MyP + __umul24(i,PPerBlock); 
	my_sts.x += s_r[j].x;
	my_sts.y += s_r[j].y;
	my_sts.z += s_r[j].z;
	my_sts.w += s_r[j].w;
      }
      if(initialize)
	sts[MyGP] = my_sts;
      else
	sts[MyGP] = Plus_Float4(my_sts, total_sts);
    }
  } // END if(STR)
  
  if(CONFT){
    
    __syncthreads();  
    s_r[MyP+offset] = my_misc;
    __syncthreads();  
    
    if( MyT == 0 ){
      for ( int i=1; i < TPerPart; i++ ){
	int j = MyP + __umul24(i,PPerBlock); 
	my_misc.x += s_r[j].x;
      }
      if(initialize)
	misc[MyGP] = my_misc;
      else
	misc[MyGP] = Plus_Float4(my_misc, total_misc);
    }
  }
} // END calc_f


template<int STR, int CONFT, bool initialize, class P, class S> __global__ void md_calc_f( P* Pot, float4* r, float4* f, float4* w, float4* sts, 
											   float4* misc, S* simBox, float* simBoxPointer, bool isCubic, 
											   NBTYPEv* NbList, float pot_factor_over2, unsigned int objectIdx ){
  extern __shared__ float4 s_r[];

  // Load the simulation box in local memory to avoid bank conflicts.
  float array[simulationBoxSize];
  simBox->loadMemory( array, simBoxPointer, isCubic );
  
  // One thread per particle reads the position
  if(MyT == 0){
    s_r[MyP] = r[MyGP];
  }

  __syncthreads();
  
  float4 my_r = s_r[MyP]; 
  
  __syncthreads(); 

  // Calculate and save force on this particle
  calc_f<STR, CONFT, initialize>( Pot, my_r, r, f, w, sts, misc, __float_as_int(my_r.w), simBox, array, NbList, pot_factor_over2, objectIdx );
}
